package com.swagger.rest.controllers

import com.swagger.rest.models.Comment
import com.swagger.rest.models.CommentInput
import com.swagger.rest.models.CommentOutput
import com.swagger.rest.repositories.BugRepository
import com.swagger.rest.repositories.CommentRepository
import com.swagger.rest.repositories.UserRepository
import org.springframework.data.domain.PageRequest
import org.springframework.data.domain.Pageable
import org.springframework.data.domain.Sort
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.web.bind.annotation.*
import java.text.SimpleDateFormat
import java.util.*

@RestController
@RequestMapping("/api/v1")
class CommentController(
    private val commentRepository: CommentRepository,
    private val bugRepository: BugRepository,
    private val userRepository: UserRepository
) {

    private fun getSortDirection(direction: String): Sort.Direction {
        if (direction == "asc") {
            return Sort.Direction.ASC
        } else if (direction == "desc") {
            return Sort.Direction.DESC
        }
        return Sort.Direction.ASC
    }

    @GetMapping("/bugs/{id}/comment")
    fun getCommentByBug(
        @RequestParam(defaultValue = 0.toString()) page: Int,
        @RequestParam(defaultValue = 3.toString()) limit: Int,
        @RequestParam(defaultValue = "comment_id, desc") sort: Array<String>,
        @PathVariable("id") id: Long
    ): Any {
        return try {
            val orders: MutableList<Sort.Order> = ArrayList()
            val column = listOf(
                "comment_id", "bug", "creator", "created", "content"
            )
            val sort2 = if (!sort.contains(",")) {
                sort + ",desc"
            } else {
                sort
            }
            val found = bugRepository.findById(id)
            if (!column.contains(sort2[0])) {
                ResponseEntity<CommentOutput>(HttpStatus.BAD_REQUEST)
            } else if (found.isEmpty) {
                ResponseEntity<CommentOutput>(HttpStatus.NOT_FOUND)
            } else {
                if (sort2[0].contains(",")) {
                    // will sort more than 2 fields
                    // sortOrder="field, direction"
                    for (sortOrder in sort2) {
                        val _sort = sortOrder.split(",".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                        orders.add(Sort.Order(getSortDirection(_sort[1]), _sort[0]))
                    }
                } else {
                    // sort=[field, direction]
                    orders.add(Sort.Order(getSortDirection(sort2[1]), sort2[0]))
                }
                val pagingSort: Pageable = PageRequest.of(page, limit, Sort.by(orders))
                val comments = commentRepository.findByBug2(id.toString(), pagingSort)
                val output = comments.map {
                    CommentOutput(
                        id = it.id,
                        bug = it.bug!!.description,
                        creator = it.creator!!.username,
                        created = SimpleDateFormat("dd-MMM-yy HH:mm:ss").format(it.created),
                        content = it.content
                    )
                }
                val ret = output.content
                val response: MutableMap<String, Any> = HashMap()
                response["currentPage"] = output.number
                response["totalRecord"] = output.totalElements
                response["totalPage"] = output.totalPages
                response["results"] = ret
                if (ret.isNotEmpty()) {
                    ResponseEntity(response, HttpStatus.OK)
                } else {
                    arrayOf<String>()
                }
            }
        } catch (e: Exception) {
            e.printStackTrace()
            ResponseEntity(e, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @GetMapping("/comments/{id}")
    fun getCommentById(@PathVariable("id") id: Long): ResponseEntity<CommentOutput> {
        val commentData = commentRepository.findById(id)
        val output = commentData.map {
            CommentOutput(
                id = it.id,
                bug = it.bug!!.description,
                creator = it.creator!!.username,
                created = SimpleDateFormat("dd-MMM-yy HH:mm:ss").format(it.created),
                content = it.content
            )
        }
        return if (commentData.isPresent) {
            ResponseEntity<CommentOutput>(output.get(), HttpStatus.OK)
        } else {
            ResponseEntity<CommentOutput>(HttpStatus.NOT_FOUND)
        }
    }

    @PostMapping("/bugs/{id}/comment")
    fun addComment(@RequestBody commentInput: CommentInput, @PathVariable("id") id: Long): Any {
        return try {
            val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
            val found = bugRepository.findById(id)
            if  (commentInput.content.isEmpty()){
                ResponseEntity(HttpStatus.BAD_REQUEST)
            } else if (commentInput.content.length > 500) {
                ResponseEntity(HttpStatus.PAYLOAD_TOO_LARGE)
            } else {
                if (found.isPresent) {
                    val saveComment = Comment()
                    saveComment.bug = found.get()
                    saveComment.creator = userId
                    saveComment.created = Date()
                    saveComment.content = commentInput.content.trim()
                    val save = commentRepository.save(saveComment)
                    if (listOf(save).isNotEmpty()) {
                        val output = CommentOutput(
                            id = save.id,
                            bug = save.bug!!.description,
                            creator = save.creator!!.username,
                            created = SimpleDateFormat("dd-MM-yyyy HH:mm:ss").format(save.created),
                            content = save.content
                        )
                        ResponseEntity(output, HttpStatus.CREATED)
                    } else {
                        ResponseEntity(null, HttpStatus.INTERNAL_SERVER_ERROR)
                    }
                } else {
                    ResponseEntity(HttpStatus.NOT_FOUND)
                }
            }
        } catch (e: Exception) {
            e.printStackTrace()
            ResponseEntity(e, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @DeleteMapping("/comments/{id}")
    fun deleteCommentById(@PathVariable("id") id: Long): Any {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        return try {
            val find = commentRepository.findById(id)
            val ownerAdmin = commentRepository.validOwnerAdmin(userId.id.toString(), id)
            if (find.isPresent) {
                if (ownerAdmin > 0) {
                    commentRepository.deleteById(id)
                    ResponseEntity(HttpStatus.OK)
                } else {
                    ResponseEntity(HttpStatus.FORBIDDEN)
                }
            } else {
                ResponseEntity(HttpStatus.NOT_FOUND)
            }
        } catch (e: Exception) {
            e.printStackTrace()
            ResponseEntity(e, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

}