package com.swagger.rest.controllers

import com.swagger.rest.models.*
import com.swagger.rest.models.Enum
import com.swagger.rest.repositories.MemberRepository
import com.swagger.rest.repositories.ProjectRepository
import com.swagger.rest.repositories.UserRepository
import jakarta.persistence.criteria.Predicate
import org.springframework.data.domain.PageRequest
import org.springframework.data.domain.Pageable
import org.springframework.data.domain.Sort
import org.springframework.data.jpa.domain.Specification
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.web.bind.annotation.*

@RestController
@RequestMapping("/api/v1")
class MemberController(
    private val memberRepository: MemberRepository,
    private val projectRepository: ProjectRepository,
    private val userRepository: UserRepository
) {
    private fun getSortDirection(direction: String): Sort.Direction {
        if (direction == "asc") {
            return Sort.Direction.ASC
        } else if (direction == "desc") {
            return Sort.Direction.DESC
        }
        return Sort.Direction.ASC
    }

    @GetMapping("/projects/{id}/member")
    fun getMemberByProject(
        @PathVariable("id") id: Long,
        @RequestParam(defaultValue = 0.toString()) page: Int,
        @RequestParam(defaultValue = 3.toString()) limit: Int,
        @RequestParam(defaultValue = "id, desc") sort: Array<String>,
        @RequestParam(required = false) role: String?
    ): Any {
        return try {
            val orders: MutableList<Sort.Order> = ArrayList()
            val column = listOf("id", "role", "project", "user")
            val sort2 = if (!sort.contains(",")) {
                sort + ",desc"
            } else {
                sort
            }
            if (!column.contains(sort2[0])) {
                ResponseEntity<ProjectMember>(HttpStatus.BAD_REQUEST)
            } else {
                if (sort[0].contains(",")) {
                    // will sort more than 2 fields
                    // sortOrder="field, direction"
                    for (sortOrder in sort2) {
                        val _sort = sortOrder.split(",".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                        orders.add(Sort.Order(getSortDirection(_sort[1]), _sort[0]))
                    }
                } else {
                    // sort=[field, direction]
                    orders.add(Sort.Order(getSortDirection(sort2[1]), sort2[0]))
                }
                val pagingSort: Pageable = PageRequest.of(page, limit, Sort.by(orders as List<Sort.Order>))
                val spec = Specification<ProjectMember> { root, query, builder ->
                    val list: MutableList<Predicate> = mutableListOf()
                    list.add(builder.equal(root.get<Project>("project").get<Long>("id"), id))
                    role?.let {
                        val r = if (Enum.Member.values()
                                .any { it.name == role.uppercase() }
                        ) Enum.Member.valueOf(role.uppercase()).ordinal else Enum.Member.UNDEFINED.ordinal
                        list.add(builder.equal(root.get<Int>("role"), r))
                    }
                    builder.and(*list.toTypedArray())
                }
                val memberData = memberRepository.findAll(spec, pagingSort)
                val output = memberData.map {
                    MemberOutput(
                        id = it.id,
                        project = it.project!!.description,
                        user = it.user!!.name,
                        role = Enum.Member.values()[it.role].name
                    )
                }
                val ret: List<MemberOutput> = output.content
                val response: MutableMap<String, Any> = HashMap()
                response["currentPage"] = output.number
                response["totalRecord"] = output.totalElements
                response["totalPage"] = output.totalPages
                response["results"] = ret
                if (ret.isNotEmpty()) {
                    ResponseEntity(response, HttpStatus.OK)
                } else {
                    arrayOf<String>()
                }
            }
        } catch (e: Exception) {
            e.printStackTrace()
            ResponseEntity(e, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @GetMapping("/projects/member/{id}")
    fun getMemberProject(@PathVariable("id") id: Long): Any {
        val memberData = memberRepository.findById(id)
        val output = memberData.map {
            MemberOutput(
                id = it.id,
                project = it.project!!.description,
                user = it.user!!.name,
                role = Enum.Member.values()[it.role].name
            )
        }
        return if (memberData.isPresent) {
            ResponseEntity<MemberOutput>(output.get(), HttpStatus.OK)
        } else {
            arrayOf<String>()
        }
    }

    @DeleteMapping("/projects/member/{id}")
    fun deleteProjectMember(@PathVariable("id") id: Long): ResponseEntity<HttpStatus> {
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        return try {
            val find = memberRepository.findById(id)
            if (find.isPresent) {
                val validOwner = projectRepository.validOwner(find.get().project!!.id.toString(), userId.id.toString())
                val validAdmin =
                    memberRepository.validRole(find.get().project!!.id.toString(), userId.id.toString(), "2")
                val targetNotAdmin = memberRepository.targetRole(
                    find.get().project!!.id.toString(), find.get().user!!.id.toString(), find.get().role.toString()
                )
                if (validOwner > 0 || (validAdmin > 0 && targetNotAdmin > 0)) {
                    memberRepository.deleteById(id)
                    ResponseEntity(HttpStatus.OK)
                } else {
                    ResponseEntity(HttpStatus.FORBIDDEN)
                }
            } else {
                ResponseEntity(HttpStatus.NOT_FOUND)
            }
        } catch (e: Exception) {
            ResponseEntity(HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @PostMapping("/projects/{id}/member")
    fun addProjectMember(
        @PathVariable("id") id: Long, @RequestBody member: MemberInput
    ): Any {
        return try {
            val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
            val found = projectRepository.findById(id)
            val user = userRepository.findById(member.user_id!!.toLong())
            val validOwner = projectRepository.validOwner(id.toString(), userId.id.toString())
            val validAdmin = memberRepository.validRole(id.toString(), userId.id.toString(), "2")
            if (found.isPresent && user.isPresent) {
                if (validOwner > 0 || validAdmin > 0) {
                    if (enumValues<Enum.Member>().any { it.name == member.role!!.uppercase() }) {
                        val role = Enum.Member.valueOf(member.role!!.uppercase()).ordinal
                        val dup =
                            memberRepository.findDup(id.toString(), role.toString(), member.user_id.toString()).size
                        if (dup > 0) {//duplicate
                            ResponseEntity<ProjectMember>(HttpStatus.CONFLICT)
                        } else {
                            val saveMember = ProjectMember()
                            saveMember.project = found.get()
                            saveMember.user = user.get()
                            saveMember.role = role
                            val save = memberRepository.save(saveMember)
                            if (listOf(save).isNotEmpty()) {
                                val output = MemberOutput(
                                    id = save.id,
                                    project = save.project!!.description,
                                    user = save.user!!.name,
                                    role = Enum.Member.values()[save.role].name
                                )
                                ResponseEntity(output, HttpStatus.CREATED)
                            } else {
                                ResponseEntity(null, HttpStatus.INTERNAL_SERVER_ERROR)
                            }
                        }
                    } else {// invalid
                        ResponseEntity<ProjectMember>(HttpStatus.BAD_REQUEST)
                    }
                } else {
                    ResponseEntity<ProjectMember>(HttpStatus.FORBIDDEN)
                }
            } else {
                ResponseEntity<ProjectMember>(HttpStatus.NOT_FOUND)
            }
        } catch (e: Exception) {
            ResponseEntity<ProjectMember>(null, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }
}