package com.swagger.rest.controllers

import com.swagger.rest.models.User
import com.swagger.rest.models.UserInput
import com.swagger.rest.repositories.MemberRepository
import com.swagger.rest.repositories.UserRepository
import jakarta.servlet.http.HttpServletResponse
import org.springframework.data.domain.PageRequest
import org.springframework.data.domain.Pageable
import org.springframework.data.domain.Sort
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.web.bind.annotation.*
import org.supercsv.io.CsvBeanWriter
import org.supercsv.io.ICsvBeanWriter
import org.supercsv.prefs.CsvPreference
import java.io.IOException
import java.text.DateFormat
import java.text.SimpleDateFormat
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashMap


@RestController
@RequestMapping("/api/v1")
class UserController(
    private val userRepository: UserRepository,
    private val passwordEncoder: PasswordEncoder,
    private val memberRepository: MemberRepository
) {

    private fun getSortDirection(direction: String): Sort.Direction {
        if (direction == "asc") {
            return Sort.Direction.ASC
        } else if (direction == "desc") {
            return Sort.Direction.DESC
        }
        return Sort.Direction.ASC
    }

    @GetMapping("/users")
    fun getUser(
        @RequestParam(defaultValue = 0.toString()) page: Int,
        @RequestParam(defaultValue = 3.toString()) limit: Int,
        @RequestParam(defaultValue = "id, desc") sort: Array<String>
    ): Any {
        return try {
            val orders: MutableList<Sort.Order> = ArrayList()
            val column = listOf("id", "name", "username")
            val sort2 = if (!sort.contains(",")) {
                sort + ",desc"
            } else {
                sort
            }
            if (!column.contains(sort2[0])) {
                ResponseEntity<User>(HttpStatus.BAD_REQUEST)
            } else {
                if (sort2[0].contains(",")) {
                    // will sort more than 2 fields
                    // sortOrder="field, direction"
                    for (sortOrder in sort2) {
                        val _sort = sortOrder.split(",".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                        orders.add(Sort.Order(getSortDirection(_sort[1]), _sort[0]))
                    }
                } else {
                    // sort=[field, direction]
                    orders.add(Sort.Order(getSortDirection(sort2[1]), sort2[0]))
                }
                val pagingSort: Pageable = PageRequest.of(page, limit, Sort.by(orders as List<Sort.Order>))
                val users = userRepository.findAll(pagingSort)
                val ret: List<User?> = users.content
                val response: MutableMap<String, Any> = HashMap()
                response["currentPage"] = users.number
                response["totalRecord"] = users.totalElements
                response["totalPage"] = users.totalPages
                response["results"] = ret
                if (ret.isNotEmpty()) {
                    ResponseEntity(response, HttpStatus.OK)
                } else {
                    arrayOf<String>()
                }
            }
        } catch (e: Exception) {
            ResponseEntity(null, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @GetMapping("/users/{id}")
    fun getUserById(@PathVariable("id") id: Long): ResponseEntity<User> {
        val userData = userRepository.findById(id)
        return if (userData.isPresent) {
            ResponseEntity<User>(userData.get(), HttpStatus.OK)
        } else {
            ResponseEntity<User>(HttpStatus.NOT_FOUND)
        }
    }

    @PostMapping("/users")
    fun addUser(@RequestBody user: UserInput): ResponseEntity<User> {
        return try {
            val found = userRepository.findByUsername(user.username!!).size
            if (user.username!!.isNotBlank()) {
                if (user.username!!.length > 100 || user.password!!.length > 100 || user.name!!.length > 255) {
                    ResponseEntity<User>(HttpStatus.PAYLOAD_TOO_LARGE)
                } else if (found > 0) {
                    ResponseEntity<User>(HttpStatus.CONFLICT)
                } else {
                    val saveUser = User()
                    saveUser.username = user.username!!.trim()
                    saveUser.password = passwordEncoder.encode(user.password)
                    saveUser.name = user.name!!
                    ResponseEntity<User>(userRepository.save(saveUser), HttpStatus.CREATED)
                }
            } else {
                ResponseEntity<User>(HttpStatus.BAD_REQUEST)
            }
        } catch (e: Exception) {
            ResponseEntity<User>(null, HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @PutMapping("/users/{id}")
    fun updateUserById(@PathVariable("id") id: Long, @RequestBody user: User): ResponseEntity<out Any> {
        val userData = userRepository.findById(id)
        val found = userRepository.findByUsername(user.username).size
        return if (user.username.isNotBlank()) {
            if (user.username.length > 100 || user.name.length > 255) {
                ResponseEntity<User>(HttpStatus.PAYLOAD_TOO_LARGE)
            } else if (userData.isPresent) {
                if ((user.username == userData.get().username && found > 0) || (user.username !== userData.get().username && found == 0)) {
                    val saveUser = userData.get()
                    saveUser.username = user.username.trim()
                    saveUser.name = user.name
                    ResponseEntity<Any>(userRepository.save(saveUser), HttpStatus.OK)
                } else {
                    ResponseEntity<User>(HttpStatus.CONFLICT)
                }
            } else {
                ResponseEntity<User>(HttpStatus.NOT_FOUND)
            }
        } else {
            ResponseEntity<User>(HttpStatus.BAD_REQUEST)
        }
    }

    @DeleteMapping("/users/{id}")
    fun deleteUserById(@PathVariable("id") id: Long): ResponseEntity<HttpStatus> {
        val find = userRepository.findById(id)
        val used = memberRepository.findByUser(id.toString()).size
        val userId = userRepository.getUserByUsername(SecurityContextHolder.getContext().authentication.name)
        val validAdmin = memberRepository.validRole("", userId.id.toString(), "2")
        val targetRole = memberRepository.validRole("", id.toString(), "2")
        val notSelf = userRepository.countByUser(id.toString(), SecurityContextHolder.getContext().authentication.name)
        return try {
            if (notSelf == 0) {
                if (validAdmin > 0 && targetRole > 0) {
                    ResponseEntity(HttpStatus.FORBIDDEN)
                } else if (used > 0) {
                    ResponseEntity(HttpStatus.RESET_CONTENT)
                } else if (find.isPresent) {
                    userRepository.deleteById(id)
                    ResponseEntity(HttpStatus.OK)
                } else {
                    ResponseEntity(HttpStatus.NOT_FOUND)
                }
            } else {
                ResponseEntity(HttpStatus.FORBIDDEN)
            }
        } catch (e: Exception) {
            ResponseEntity(HttpStatus.INTERNAL_SERVER_ERROR)
        }
    }

    @PutMapping("/users/{id}/password")
    fun updatePassword(@PathVariable("id") id: Long, @RequestBody userInput: UserInput): ResponseEntity<out Any> {
        val userData = userRepository.findById(id)
        return if (userInput.newPassword!!.isNotBlank() || userInput.oldPassword!!.isNotBlank()) {
            if (userInput.newPassword!!.length > 100) {
                ResponseEntity<User>(HttpStatus.PAYLOAD_TOO_LARGE)
            } else if (userData.isPresent) {
                if (passwordEncoder.matches(userInput.oldPassword, userData.get().password)) {
                    val saveUser = userData.get()
                    saveUser.password = passwordEncoder.encode(userInput.newPassword)
                    ResponseEntity<Any>(userRepository.save(saveUser), HttpStatus.OK)
                } else {
                    ResponseEntity<User>(HttpStatus.FORBIDDEN)
                }
            } else {
                ResponseEntity<User>(HttpStatus.NOT_FOUND)
            }
        } else {
            ResponseEntity<User>(HttpStatus.BAD_REQUEST)
        }
    }

}